
From: Matt Tolentino <metolent@snoqualmie.dp.intel.com>

This patch adds in the necessary support for sparsemem such that x86-64
kernels may use sparsemem as an alternative to discontigmem for NUMA
kernels.  Note that this does no preclude one from continuing to build NUMA
kernels using discontigmem, but merely allows the option to build NUMA
kernels with sparsemem.

Interestingly, the use of sparsemem in lieu of discontigmem in NUMA kernels
results in reduced text size for otherwise equivalent kernels as shown in
the example builds below:

   text	   data	    bss	    dec	    hex	filename
2371036	 765884	1237108	4374028	 42be0c	vmlinux.discontig
2366549	 776484	1302772	4445805	 43d66d	vmlinux.sparse

Signed-off-by: Matt Tolentino <matthew.e.tolentino@intel.com>
Signed-off-by: Dave Hansen <haveblue@us.ibm.com>
Signed-off-by: Andrew Morton <akpm@osdl.org>
---

 arch/x86_64/kernel/setup.c     |   11 +++++++++--
 arch/x86_64/mm/numa.c          |    8 ++++++++
 include/asm-x86_64/bitops.h    |    2 --
 include/asm-x86_64/sparsemem.h |   26 ++++++++++++++++++++++++++
 4 files changed, 43 insertions(+), 4 deletions(-)

diff -puN arch/x86_64/kernel/setup.c~add-x86-64-specific-support-for-sparsemem arch/x86_64/kernel/setup.c
--- 25/arch/x86_64/kernel/setup.c~add-x86-64-specific-support-for-sparsemem	Wed May 18 14:46:20 2005
+++ 25-akpm/arch/x86_64/kernel/setup.c	Wed May 18 14:46:37 2005
@@ -40,6 +40,8 @@
 #include <linux/acpi.h>
 #include <linux/kallsyms.h>
 #include <linux/edd.h>
+#include <linux/mmzone.h>
+
 #include <asm/mtrr.h>
 #include <asm/uaccess.h>
 #include <asm/system.h>
@@ -378,9 +380,12 @@ static __init void parse_cmdline_early (
 }
 
 #ifndef CONFIG_NUMA
-static void __init contig_initmem_init(void)
+static void __init
+contig_initmem_init(unsigned long start_pfn, unsigned long end_pfn)
 {
         unsigned long bootmap_size, bootmap; 
+
+	memory_present(0, start_pfn, end_pfn);
         bootmap_size = bootmem_bootmap_pages(end_pfn)<<PAGE_SHIFT;
         bootmap = find_e820_area(0, end_pfn<<PAGE_SHIFT, bootmap_size);
         if (bootmap == -1L) 
@@ -557,7 +562,7 @@ void __init setup_arch(char **cmdline_p)
 #ifdef CONFIG_NUMA
 	numa_initmem_init(0, end_pfn); 
 #else
-	contig_initmem_init(); 
+	contig_initmem_init(0, end_pfn);
 #endif
 
 	/* Reserve direct mapping */
@@ -618,6 +623,8 @@ void __init setup_arch(char **cmdline_p)
 		}
 	}
 #endif
+
+	sparse_init();
 	paging_init();
 
 	check_ioapic();
diff -puN arch/x86_64/mm/numa.c~add-x86-64-specific-support-for-sparsemem arch/x86_64/mm/numa.c
--- 25/arch/x86_64/mm/numa.c~add-x86-64-specific-support-for-sparsemem	Wed May 18 14:46:20 2005
+++ 25-akpm/arch/x86_64/mm/numa.c	Wed May 18 14:46:20 2005
@@ -66,6 +66,13 @@ int __init compute_hash_shift(struct nod
 	return -1; 
 }
 
+#ifdef CONFIG_SPARSEMEM
+int early_pfn_to_nid(unsigned long pfn)
+{
+	return phys_to_nid(pfn << PAGE_SHIFT);
+}
+#endif
+
 /* Initialize bootmem allocator for a node */
 void __init setup_node_bootmem(int nodeid, unsigned long start, unsigned long end)
 { 
@@ -80,6 +87,7 @@ void __init setup_node_bootmem(int nodei
 	start_pfn = start >> PAGE_SHIFT;
 	end_pfn = end >> PAGE_SHIFT;
 
+	memory_present(nodeid, start_pfn, end_pfn);
 	nodedata_phys = find_e820_area(start, end, pgdat_size); 
 	if (nodedata_phys == -1L) 
 		panic("Cannot find memory pgdat in node %d\n", nodeid);
diff -puN include/asm-x86_64/bitops.h~add-x86-64-specific-support-for-sparsemem include/asm-x86_64/bitops.h
--- 25/include/asm-x86_64/bitops.h~add-x86-64-specific-support-for-sparsemem	Wed May 18 14:46:20 2005
+++ 25-akpm/include/asm-x86_64/bitops.h	Wed May 18 14:46:20 2005
@@ -411,8 +411,6 @@ static __inline__ int ffs(int x)
 /* find last set bit */
 #define fls(x) generic_fls(x)
 
-#define ARCH_HAS_ATOMIC_UNSIGNED 1
-
 #endif /* __KERNEL__ */
 
 #endif /* _X86_64_BITOPS_H */
diff -puN /dev/null include/asm-x86_64/sparsemem.h
--- /dev/null	Thu Apr 11 07:25:15 2002
+++ 25-akpm/include/asm-x86_64/sparsemem.h	Wed May 18 14:46:20 2005
@@ -0,0 +1,26 @@
+#ifndef _ASM_X86_64_SPARSEMEM_H
+#define _ASM_X86_64_SPARSEMEM_H 1
+
+#ifdef CONFIG_SPARSEMEM
+
+/*
+ * generic non-linear memory support:
+ *
+ * 1) we will not split memory into more chunks than will fit into the flags
+ *    field of the struct page
+ *
+ * SECTION_SIZE_BITS		2^n: size of each section
+ * MAX_PHYSADDR_BITS		2^n: max size of physical address space
+ * MAX_PHYSMEM_BITS		2^n: how much memory we can have in that space
+ *
+ */
+
+#define SECTION_SIZE_BITS	27 /* matt - 128 is convenient right now */
+#define MAX_PHYSADDR_BITS	40
+#define MAX_PHYSMEM_BITS	40
+
+extern int early_pfn_to_nid(unsigned long pfn);
+
+#endif /* CONFIG_SPARSEMEM */
+
+#endif /* _ASM_X86_64_SPARSEMEM_H */
_
